/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Call Campon facility
 * 
 * Copyright (C) 1999, Mark Spencer
 * Copyright (C) 2003, Paul Liew
 *
 * Mark Spencer <markster@linux-support.net>
 * Paul Liew <pliew@atp.org.au>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/options.h>
#include <asterisk/module.h>
#include <asterisk/translate.h>
#include <asterisk/say.h>
#include <asterisk/parking.h>
#include <asterisk/musiconhold.h>
#include <asterisk/cli.h>
#include <asterisk/config.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <netinet/in.h>

#include <pthread.h>

#define DEFAULT_CAMPON_MOH	"default"
#define DEFAULT_CAMPON_RECALL	45
#define DEFAULT_CAMPON_DIGIT_TIMEOUT	5

#define DEFAULT_RETRY		5
#define DEFAULT_TIMEOUT		15
#define RECHECK				1		/* Recheck every second to see we we're at the top yet */

static	char	campon_moh[80];
static	int		campon_recall;
static	int		campon_digit_timeout;

static char *tdesc = "Call Campon";

/* Campon desriptions */
static char *campon_app = "Campon";
static char *campon_synopsis = "Camp a call for an extension";
static char *campon_descrip =
"  Campon(technology/number[|options):\n"
"Camps an incoming call onta a particular extension.\n"
"  This application returns -1 if the originating channel hangs up, or if the\n"
"call is bridged and  either of the parties in the bridge terminate the call.\n"
"Returns 0 if caller exists by pressing 1 or 2.\n"
"The option string may contain zero or more of the following characters:\n"
"      't' -- allow the called user transfer the calling user\n"
"      'T' -- to allow the calling user to transfer the call.\n"
"      'H' -- to allow the calling user to hangup the call by pressing *.\n";


/* We define a customer "local user" structure because we
   use it not only for keeping track of what is in use but
   also for keeping track of who we're dialing. */

struct localuser {
	struct ast_channel *chan;
	char numsubst[256];
	char tech[40];
	int stillgoing;
	int allowredirect_in;
	int allowredirect_out;
	int allowdisconnect;
	struct member *member;
	struct localuser *next;
};

LOCAL_USER_DECL;

struct queue_ent {
	struct ast_campon_queue *parent;	/* What queue is our parent */
	char context[80];		/* Context when user exits queue */
	int pos;					/* Where we are in the queue */
	time_t recall;				/* recall time for campon message */
	struct ast_channel *chan;	/* Our channel */
	struct queue_ent *next;		/* The next queue entry */
};

struct member {
	char tech[80];				/* Technology */
	char loc[256];				/* Location */
};

struct ast_campon_queue {
	ast_mutex_t	lock;	
	char name[80];			/* Name of the queue */
	char context[80];		/* Announcement to play */
	int count;				/* How many entries are in the queue */
	int maxlen;				/* Max number of entries in queue */

	int dead;				/* Whether this queue is dead or not */
	int retry;				/* Retry calling everyone after this amount of time */
	int timeout;			/* How long to wait for an answer */
	
	struct member *members;	/* Member channels to be tried */
	struct queue_ent *head;	/* Start of the actual queue */
	struct ast_campon_queue *next;	/* Next call queue */
};

static struct ast_campon_queue *queues = NULL;
static ast_mutex_t qlock = AST_MUTEX_INITIALIZER;

static int join_queue(char *queuename, struct queue_ent *qe)
{
	struct ast_campon_queue *q;
	struct queue_ent *cur, *prev = NULL;
	int res = -1;
	int pos = 0;
	ast_mutex_lock(&qlock);
	q = queues;
	while(q) {
		if (!strcasecmp(q->name, queuename)) {
			/* This is our one */
			ast_mutex_lock(&q->lock);
			if (q->members && (!q->maxlen || (q->count < q->maxlen))) {
				/* There's space for us, put us at the end */
				prev = NULL;
				cur = q->head;
				while(cur) {
					cur->pos = ++pos;
					prev = cur;
					cur = cur->next;
				}
				if (prev)
					prev->next = qe;
				else
					q->head = qe;
				/* Fix additional pointers and
				  information  */
				qe->next = NULL;
				qe->parent = q;
				qe->pos = ++pos;
				strncpy(qe->context, q->context, sizeof(qe->context));
				q->count++;
				res = 0;
			}
			ast_mutex_unlock(&q->lock);
			break;
		}
		q = q->next;
	}
	ast_mutex_unlock(&qlock);
	return res;
}

static void destroy_queue(struct ast_campon_queue *q)
{
	struct ast_campon_queue *cur, *prev = NULL;
	ast_mutex_lock(&qlock);
	cur = queues;
	while(cur) {
		if (cur == q) {
			if (prev)
				prev->next = cur->next;
			else
				queues = cur->next;
		} else {
			prev = cur;
		}
		cur = cur->next;
	}
	ast_mutex_unlock(&qlock);
	free(q);
}

static void leave_queue(struct queue_ent *qe)
{
	struct ast_campon_queue *q;
	struct queue_ent *cur, *prev = NULL;
	int pos = 0;
	q = qe->parent;
	if (!q)
		return;
	ast_mutex_lock(&q->lock);

	prev = NULL;
	cur = q->head;
	while(cur) {
		if (cur == qe) {
			q->count--;

			/* Take us out of the queue */
			if (prev)
				prev->next = cur->next;
			else
				q->head = cur->next;
		} else {
			cur->pos = ++pos;
			prev = cur;
		}
		cur = cur->next;
	}
	ast_mutex_unlock(&q->lock);
	if (q->dead && !q->count) {	
		/* It's dead and nobody is in it, so kill it */
		destroy_queue(q);
	}
}

static void hanguptree(struct localuser *outgoing, struct ast_channel *exception)
{
	/* Hang up a tree of stuff */
	struct localuser *oo;
	while(outgoing) {
		/* Hangup any existing lines we have open */
		if (outgoing->chan && (outgoing->chan != exception))
			ast_hangup(outgoing->chan);
		oo = outgoing;
		outgoing=outgoing->next;
		free(oo);
	}
}

static int ring_entry(struct queue_ent *qe, struct localuser *tmp)
{
	int res;
	/* Request the peer */
	tmp->chan = ast_request(tmp->tech, qe->chan->nativeformats, tmp->numsubst);
	if (!tmp->chan) {			/* If we can't, just go on to the next call */
#if 0
		ast_log(LOG_NOTICE, "Unable to create channel of type '%s'\n", cur->tech);
#endif			
		if (qe->chan->cdr)
			ast_cdr_busy(qe->chan->cdr);
		tmp->stillgoing = 0;
		return 0;
	}
	tmp->chan->appl = "AppCampon";
	tmp->chan->data = "(Outgoing Line)";
	tmp->chan->whentohangup = 0;
	if (tmp->chan->callerid)
		free(tmp->chan->callerid);
	if (tmp->chan->ani)
		free(tmp->chan->ani);
	if (qe->chan->callerid)
		tmp->chan->callerid = strdup(qe->chan->callerid);
	else
		tmp->chan->callerid = NULL;
	if (qe->chan->ani)
		tmp->chan->ani = strdup(qe->chan->ani);
	else
		tmp->chan->ani = NULL;
	/* Presense of ADSI CPE on outgoing channel follows ours */
	tmp->chan->adsicpe = qe->chan->adsicpe;
	/* Place the call, but don't wait on the answer */
	res = ast_call(tmp->chan, tmp->numsubst, 0);
	if (res) {
		/* Again, keep going even if there's an error */
		if (option_debug)
			ast_log(LOG_DEBUG, "ast call on peer returned %d\n", res);
		else if (option_verbose > 2)
			ast_verbose(VERBOSE_PREFIX_3 "Couldn't call %s\n", tmp->numsubst);
		ast_hangup(tmp->chan);
		tmp->chan = NULL;
		tmp->stillgoing = 0;
		return 0;
	} else
		if (option_verbose > 2)
			ast_verbose(VERBOSE_PREFIX_3 "Called %s\n", tmp->numsubst);
	return 0;
}

static int valid_exit(struct queue_ent *qe, char digit)
{
	char tmp[2];
	if (!strlen(qe->context))
		return 0;
	tmp[0] = digit;
	tmp[1] = '\0';
	if (ast_exists_extension(qe->chan, qe->context, tmp, 1, qe->chan->callerid)) {
		strncpy(qe->chan->context, qe->context, sizeof(qe->chan->context) - 1);
		strncpy(qe->chan->exten, tmp, sizeof(qe->chan->exten) - 1);
		qe->chan->priority = 0;
		return 1;
	}
	return 0;
}

#define MAX 256

static struct localuser *wait_for_answer(struct queue_ent *qe, struct localuser *outgoing, int *to, int *allowredir_in, int *allowredir_out, int *allowdisconnect, char *digit)
{
	char *queue = qe->parent->name;
	struct localuser *o;
	int found;
	int numlines;
	int sentringing = 0;
	int numbusies = 0;
	int orig = *to;
	struct ast_frame *f;
	struct localuser *peer = NULL;
	struct ast_channel *watchers[MAX];
	int pos;
	struct ast_channel *winner;
	struct ast_channel *in = qe->chan;
	
	while(*to && !peer) {
		o = outgoing;
		found = -1;
		pos = 1;
		numlines = 0;
		watchers[0] = in;
		while(o) {
			/* Keep track of important channels */
			if (o->stillgoing && o->chan) {
				watchers[pos++] = o->chan;
				found = 1;
			}
			o = o->next;
			numlines++;
		}
		if (found < 0) {
			if (numlines == numbusies) {
				ast_log(LOG_DEBUG, "Everyone is busy at this time\n");
			} else {
				ast_log(LOG_NOTICE, "No one is answered queue %s\n", queue);
			}
			*to = 0;
			return NULL;
		}
		winner = ast_waitfor_n(watchers, pos, to);
		o = outgoing;
		while(o) {
			if (o->stillgoing && (o->chan) &&  (o->chan->_state == AST_STATE_UP)) {
				if (!peer) {
					if (option_verbose > 2)
						ast_verbose( VERBOSE_PREFIX_3 "%s answered %s\n", o->chan->name, in->name);
					peer = o;
					*allowredir_in = o->allowredirect_in;
					*allowredir_out = o->allowredirect_out;
					*allowdisconnect = o->allowdisconnect;
				}
			} else if (o->chan && (o->chan == winner)) {
				f = ast_read(winner);
				if (f) {
					if (f->frametype == AST_FRAME_CONTROL) {
						switch(f->subclass) {
					    case AST_CONTROL_ANSWER:
							/* This is our guy if someone answered. */
							if (!peer) {
								if (option_verbose > 2)
									ast_verbose( VERBOSE_PREFIX_3 "%s answered %s\n", o->chan->name, in->name);
								peer = o;
								*allowredir_in = o->allowredirect_in;
								*allowredir_out = o->allowredirect_out;
								*allowdisconnect = o->allowdisconnect;
							}
							break;
						case AST_CONTROL_BUSY:
							if (option_verbose > 2)
								ast_verbose( VERBOSE_PREFIX_3 "%s is busy\n", o->chan->name);
							o->stillgoing = 0;
							if (in->cdr)
								ast_cdr_busy(in->cdr);
							ast_hangup(o->chan);
							o->chan = NULL;
							ring_entry(qe, outgoing);
							numbusies++;
							break;
						case AST_CONTROL_CONGESTION:
							if (option_verbose > 2)
								ast_verbose( VERBOSE_PREFIX_3 "%s is circuit-busy\n", o->chan->name);
							o->stillgoing = 0;
							if (in->cdr)
								ast_cdr_busy(in->cdr);
							ast_hangup(o->chan);
							o->chan = NULL;
							ring_entry(qe, outgoing);
							numbusies++;
							break;
						case AST_CONTROL_RINGING:
							if (option_verbose > 2)
								ast_verbose( VERBOSE_PREFIX_3 "%s is ringing\n", o->chan->name);
							if (!sentringing) {
#if 0
								ast_indicate(in, AST_CONTROL_RINGING);
#endif								
								sentringing++;
							}
							break;
						case AST_CONTROL_OFFHOOK:
							/* Ignore going off hook */
							break;
						default:
							ast_log(LOG_DEBUG, "Dunno what to do with control type %d\n", f->subclass);
						}
					}
					ast_frfree(f);
				} else {
					o->stillgoing = 0;
					ast_hangup(o->chan);
					o->chan = NULL;
					ring_entry(qe, outgoing);
				}
			}
			o = o->next;
		}
		if (winner == in) {
			f = ast_read(in);
#if 0
			if (f && (f->frametype != AST_FRAME_VOICE))
					printf("Frame type: %d, %d\n", f->frametype, f->subclass);
			else if (!f || (f->frametype != AST_FRAME_VOICE))
				printf("Hangup received on %s\n", in->name);
#endif
			if (!f || ((f->frametype == AST_FRAME_CONTROL) && (f->subclass == AST_CONTROL_HANGUP))) {
				/* Got hung up */
				*to=-1;
				return NULL;
			}
			if (f && (f->frametype == AST_FRAME_DTMF) && allowdisconnect && (f->subclass == '*')) {
			    if (option_verbose > 3)
				ast_verbose(VERBOSE_PREFIX_3 "User hit %c to disconnect call.\n", f->subclass);
				*to=0;
				return NULL;
			}
			if (f && (f->frametype == AST_FRAME_DTMF) && (f->subclass != '*') && valid_exit(qe, f->subclass)) {
				if (option_verbose > 3)
					ast_verbose(VERBOSE_PREFIX_3 "User pressed digit: %c", f->subclass);
				*to=0;
				*digit=f->subclass;
				return NULL;
			}
		}
		if (!*to && (option_verbose > 2))
			ast_verbose( VERBOSE_PREFIX_3 "Nobody picked up in %d ms\n", orig);
	}

	return peer;
	
}

static int wait_our_turn(struct queue_ent *qe)
{
	struct queue_ent *ch;
	int res = 0;
	for (;;) {
		/* Atomically read the parent head */
		ast_mutex_lock(&qe->parent->lock);
		ch = qe->parent->head;
		ast_mutex_unlock(&qe->parent->lock);
		/* If we are now at the top of the head, break out */
		if (qe->parent->head == qe)
			break;
		/* Wait a second before checking again */
		res = ast_waitfordigit(qe->chan, RECHECK * 1000);
		if (res)
			break;
	}
	return res;
}

static int try_calling(struct queue_ent *qe, char *options)
{
	struct member *cur;
	struct localuser *outgoing=NULL, *tmp = NULL;
	int to;
	int allowredir_in=0;
	int allowredir_out=0;
	int allowdisconnect=0;
	char restofit[AST_MAX_EXTENSION];
	char *newnum;
	struct ast_channel *peer;
	struct localuser *lpeer;
	int res = 0, bridge = 0;
	int zapx = 2;
	char digit = 0;
	/* Hold the lock while we setup the outgoing calls */
	ast_mutex_lock(&qe->parent->lock);
	cur = qe->parent->members;
	
	/* Get a technology/[device:]number pair */
	tmp = malloc(sizeof(struct localuser));
	if (!tmp) {
		ast_log(LOG_WARNING, "Out of memory\n");
		goto out;
	}
	memset(tmp, 0, sizeof(struct localuser));
	tmp->stillgoing = -1;
	if (options) {
		if (strchr(options, 't'))
			tmp->allowredirect_in = 1;
		if (strchr(options, 'T'))
			tmp->allowredirect_out = 1;
		if (strchr(options, 'H'))
			tmp->allowdisconnect = 1;
	}

	tmp->member = cur;		/* Never directly dereference!  Could change on reload */
	strncpy(tmp->tech, cur->tech, sizeof(tmp->tech)-1);
	strncpy(tmp->numsubst, cur->loc, sizeof(tmp->numsubst)-1);
	/* If we're dialing by extension, look at the extension to know what to dial */
	if ((newnum = strstr(tmp->numsubst, "BYEXTENSION"))) {
		strncpy(restofit, newnum + strlen("BYEXTENSION"), sizeof(restofit)-1);
		snprintf(newnum, sizeof(tmp->numsubst) - (newnum - tmp->numsubst), "%s%s", qe->chan->exten,restofit);
		if (option_debug)
			ast_log(LOG_DEBUG, "Dialing by extension %s\n", tmp->numsubst);
	}
	/* Put them in the list of outgoing thingies...  We're ready now. 
	   XXX If we're forcibly removed, these outgoing calls won't get
	   hung up XXX */
	tmp->next = outgoing;
	outgoing = tmp;		

	if (qe->parent->timeout)
		to = qe->parent->timeout * 1000;
	else
		to = -1;
	ring_entry(qe, outgoing);
	ast_mutex_unlock(&qe->parent->lock);
	lpeer = wait_for_answer(qe, outgoing, &to, &allowredir_in, &allowredir_out, &allowdisconnect, &digit);
	if (lpeer)
		peer = lpeer->chan;
	else
		peer = NULL;
	if (!peer) {
		if (to) {
			/* Musta gotten hung up */
			res = -1;
		} else {
			if (digit && valid_exit(qe, digit))
				res=digit;
			else
				/* Nobody answered, next please? */
				res=0;
		}
		goto out;
	}
	if (peer) {
		/* Ah ha!  Someone answered within the desired timeframe.  Of course after this
		   we will always return with -1 so that it is hung up properly after the 
		   conversation.  */
		if (!strcmp(qe->chan->type,"Zap")) {
			ast_channel_setoption(qe->chan,AST_OPTION_TONE_VERIFY,&zapx,sizeof(char),0);
		}			
		if (!strcmp(peer->type,"Zap")) {
			ast_channel_setoption(peer,AST_OPTION_TONE_VERIFY,&zapx,sizeof(char),0);
		}
		hanguptree(outgoing, peer);
		/* Stop music on hold */
		ast_moh_stop(qe->chan);
		outgoing = NULL;
		/* If appropriate, log that we have a destination channel */
		if (qe->chan->cdr)
			ast_cdr_setdestchan(qe->chan->cdr, peer->name);
		/* Make sure channels are compatible */
		res = ast_channel_make_compatible(qe->chan, peer);
		if (res < 0) {
			ast_log(LOG_WARNING, "Had to drop call because I couldn't make %s compatible with %s\n", qe->chan->name, peer->name);
			ast_hangup(peer);
			return -1;
		}
		/* Drop out of the queue at this point, to prepare for next caller */
		leave_queue(qe);			
		bridge = ast_bridge_call(qe->chan, peer, allowredir_in, allowredir_out, allowdisconnect);

		if(bridge != AST_PBX_NO_HANGUP_PEER)
			ast_hangup(peer);

		if( bridge == 0 ) res=1; /* JDG: bridge successfull, leave app_queue */
		else res = bridge; /* bridge error, stay in the queue */
	}	
out:
	hanguptree(outgoing, NULL);
	return res;
}

static int wait_a_bit(struct queue_ent *qe)
{
	int retrywait;
	/* Hold the lock while we setup the outgoing calls */
	ast_mutex_lock(&qe->parent->lock);
	retrywait = qe->parent->retry * 1000;
	ast_mutex_unlock(&qe->parent->lock);
	return ast_waitfordigit(qe->chan, retrywait);
}


static int campon_message(struct ast_channel *chan, char *fname)
{
	int	d;

	d = ast_streamfile(chan, fname, chan->language);
	if ( !d ) {
		d = ast_waitstream(chan, AST_DIGIT_ANY);
	}
	if ( !d ) {
		d = ast_waitfordigit(chan, campon_digit_timeout * 1000);
	}
	return d;
}

static int campon_exec(struct ast_channel *chan, void *data)
{
	int res=-1;
	int new;
	struct localuser *u;
	char *queuename;
	char info[512];
	char *options = NULL;
	struct ast_campon_queue *q;
	struct member *prev, *cur;
	char *tmp;
	char	destination[32];
	int		d;
	time_t	curr_time;
	
	/* Our queue entry */
	struct queue_ent qe;
	
	if (!data) {
		ast_log(LOG_WARNING, "CAMPON requires an argument (technology/number|options)\n");
		return -1;
	}

	/* play campon message */
	if ( chan->_state != AST_STATE_UP )
		ast_answer(chan);
	d = campon_message(chan, "campon");
	if ( d == '1' ) {
		return 0;
	}
	else if ( d == '2' && ast_exists_extension(chan, chan->context, chan->exten,
				chan->priority + 101, chan->callerid) ) {
		chan->priority += 100;
		return 0;
	}
	else if ( d == '*' ) {
		return -1;
	}
	
	LOCAL_USER_ADD(u);
	
	/* Parse our arguments XXX Check for failure XXX */
	strncpy(info, (char *)data, strlen((char *)data) + AST_MAX_EXTENSION-1);
	queuename = info;
	if (queuename) {
		options = strchr(queuename, '|');
		if (options) {
			*options = '\0';
			options++;
		}
	}
	strcpy(destination, queuename);
	/* make sure queue entry for this phone/extension exists otherwise create */
	q = queues;
	while (q) {
		if ( !strcmp(q->name, queuename) )
			break;
		q = q->next;
	}
	if (!q) {
		/* Make one then */
		q = malloc(sizeof(struct ast_campon_queue));
		if (q) {
			/* Initialize it */
			memset(q, 0, sizeof(struct ast_campon_queue));
			ast_mutex_init(&q->lock);
			strncpy(q->name, queuename, sizeof(q->name));
			new = 1;
		} else new = 0;
	} else
		new = 0;
	if (q) {
		if (!new) 
			ast_mutex_lock(&q->lock);
		/* Re-initialize the queue */
		q->dead = 0;
		q->retry = 0;
		q->timeout = -1;
		q->maxlen = 0;
		strcpy(q->context, "");
		cur = malloc(sizeof(struct member));
		if ( cur ) {
			memset(cur, 0, sizeof(struct member));
			strncpy(cur->tech, queuename, sizeof(cur->tech) - 1);
			if ((tmp = strchr(cur->tech, '/')))
				*tmp = '\0';
			if ((tmp = strchr(queuename, '/'))) {
				tmp++;
				strncpy(cur->loc, tmp, sizeof(cur->loc) - 1);
				if ((tmp = strchr(cur->loc, ',')))
					*tmp = '\0';
			}
		}
		q->members = cur;
		prev = cur;
		if (q->retry < 1)
			q->retry = DEFAULT_RETRY;
		if (q->timeout < 0)
			q->timeout = DEFAULT_TIMEOUT;
		if (q->maxlen < 0)
			q->maxlen = 0;
		if (!new) 
			ast_mutex_unlock(&q->lock);
		if (new) {
			q->next = queues;
			queues = q;
		}
	}

	/* Setup our queue entry */
	memset(&qe, 0, sizeof(qe));
	qe.chan = chan;
	if (!join_queue(destination, &qe)) {
		/* Start music on hold */
		ast_moh_start(chan, campon_moh);
		for (;;) {
			res = wait_our_turn(&qe);
			/* If they hungup, return immediately */
			if (res < 0) {
				if (option_verbose > 2) {
					ast_verbose(VERBOSE_PREFIX_3 "User disconnected while waiting their turn\n");
					res = -1;
				}
				break;
			}
			if (!res)
				break;
			if (valid_exit(&qe, res))
				break;
		}
		if (!res) {
			qe.recall = time(NULL) + campon_recall;
			for (;;) {
				res = try_calling(&qe, options);
				if (res)
					break;
				res = wait_a_bit(&qe);
				curr_time = time(NULL);
				if ( curr_time >= qe.recall ) {
					ast_moh_stop(chan);
					d = campon_message(chan, "campon-repeat");
					if ( d == '1' ) {
						res = 0;
						break;
					}
					else if ( d == '2' && ast_exists_extension(chan, 
								chan->context, chan->exten,
								chan->priority + 101, chan->callerid) ) {
						chan->priority += 100;
						return 0;
					}
					else {
						ast_moh_start(chan, campon_moh);
						qe.recall = time(NULL) + campon_recall;
					}
				}
				if (res < 0) {
					if (option_verbose > 2) {
						ast_verbose(VERBOSE_PREFIX_3 "User disconnected when they almost made it\n");
						res = -1;
					}
					break;
				}
				if (res && valid_exit(&qe, res))
					break;
			}
		}
		/* Don't allow return code > 0 */
		if (res > 0 && res != AST_PBX_KEEPALIVE) {
			res = 0;	
			ast_moh_stop(chan);
		}
		leave_queue(&qe);
	} else {
		ast_log(LOG_WARNING, "Unable to join queue '%s'\n", queuename);
		res =  0;
	}
	LOCAL_USER_REMOVE(u);
	return res;
}

static void reload_campon(void)
{
	struct ast_campon_queue *q, *ql, *qn;
	struct ast_config *cfg;
	char *cat;
	struct ast_variable *var;

	strcpy(campon_moh, DEFAULT_CAMPON_MOH);
	campon_recall = DEFAULT_CAMPON_RECALL;
	campon_digit_timeout = DEFAULT_CAMPON_DIGIT_TIMEOUT;
	cfg = ast_load("campon.conf");
	if (!cfg) {
		ast_log(LOG_NOTICE, "No campon config file, using defaults\n");
		return;
	}
	var = ast_variable_browse(cfg, "general");
	while (var) {
		if ( !strcasecmp(var->name, "music") ) {
			strncpy(campon_moh, var->value, sizeof(campon_moh) - 1);
		}
		else if ( !strcasecmp(var->name, "recall") ) {
			campon_recall = atoi(var->value);
		}
		else if ( !strcasecmp(var->name, "responsetime") ) {
			campon_digit_timeout = atoi(var->value);
		}
		else {
			ast_log(LOG_WARNING, "Unknown keyword in campon '%s' at line %d of campon.conf\n", var->name, var->lineno);
		}
		var = var->next;
	}
	ast_destroy(cfg);
	ast_mutex_lock(&qlock);
	/* Mark all queues as dead for the moment */
	q = queues;
	ql = NULL;
	while(q) {
		qn = q->next;
		if (q->dead) {
			if (ql)
				ql->next = q->next;
			else
				queues = q->next;
			if (!q->count) {
				free(q);
			} else
				ast_log(LOG_WARNING, "XXX Leaking a litttle memory :( XXX\n");
		} else
			ql = q;
		q = qn;
	}
	ast_mutex_unlock(&qlock);
}

int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
	return ast_unregister_application(campon_app);
}

int load_module(void)
{
	int res;
	res = ast_register_application(campon_app, campon_exec, campon_synopsis, campon_descrip);
	reload_campon();
	return res;
}


int reload(void)
{
	reload_campon();
	return 0;
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
